<!DOCTYPE html>
<html>

<head>
  <title>OnionShare</title>
  <meta charset="utf-8" />
  <meta name="viewport" content="width=device-width, initial-scale=1">
</head>

<body>
  <p>OnionShare download in progress</p>
</body>

</html>