<!DOCTYPE html>
<html>

<head>
  <title>{% if title %}{{ title }}{% else %}OnionShare Dropbox{% endif %}</title>
  <meta charset="utf-8" />
  <meta name="viewport" content="width=device-width, initial-scale=1">
  <link href="{{ static_url_path }}/img/favicon.ico" rel="icon" type="image/x-icon">
  <link rel="stylesheet" rel="subresource" type="text/css" href="{{ static_url_path }}/css/style.css" media="all">
</head>

<body>

  <header class="clearfix">
    <img class="logo" src="{{ static_url_path }}/img/logo.png" title="OnionShare">
    <h1>{% if title %}{{ title }}{% else %}OnionShare Dropbox{% endif %}</h1>
  </header>

  <div class="upload-wrapper">
    <p><img class="logo" src="{{ static_url_path }}/img/logo_large.png" title="OnionShare"></p>

    {% if not disable_text and not disable_files %}
    <p class="upload-header">Submit Files or Messages</p>
    <p class="upload-description">You can submit files, a message, or both.</p>
    <p class="upload-description">Remember, you are accessing this service anonymously! Provide contact info if you want a response to the message.</p>
    {% endif %}
    {% if not disable_text and disable_files %}
    <p class="upload-header">Submit Messages</p>
    <p class="upload-description">You can submit a message.</p>
    <p class="upload-description">Remember, you are accessing this service anonymously! Provide contact info if you want a response to the message.</p>
    {% endif %}
    {% if disable_text and not disable_files %}
    <p class="upload-header">Submit Files</p>
    <p class="upload-description">You can submit files</p>
    {% endif %}

    <div id="uploads"></div>

    <div>
      <ul id="flashes" class="flashes">
        {% with messages = get_flashed_messages(with_categories=true) %}
        {% if messages %}
        {% for category, message in messages %}
        <li class="{{ category }}">{{ message }}</li>
        {% endfor %}
        {% endif %}
        {% endwith %}
      </ul>
    </div>

    <form id="send" method="post" enctype="multipart/form-data" action="/upload">
      {% if not disable_files %}
      <p><input type="file" id="file-select" name="file[]" multiple /></p>
      {% endif %}
      {% if not disable_text %}
      <p><textarea id="text" name="text" placeholder="Write a message (max length 524288 characters)" maxlength="524288"></textarea></p>
      {% endif %}
      <p><button type="submit" id="send-button" class="button">Submit</button></p>
    </form>

  </div>
  <script src="{{ static_url_path }}/js/jquery-3.5.1.min.js"></script>
  <script async src="{{ static_url_path }}/js/receive.js" id="receive-script"></script>
</body>

</html>
