<!DOCTYPE html>
<html>

<head>
  <title>OnionShare is closed</title>
  <meta charset="utf-8" />
  <meta name="viewport" content="width=device-width, initial-scale=1">
  <link href="{{ static_url_path }}/img/favicon.ico" rel="icon" type="image/x-icon">
  <link rel="stylesheet" rel="subresource" type="text/css" href="{{ static_url_path }}/css/style.css" media="all">
</head>

<body>
  <header class="clearfix">
    <img class="logo" src="{{ static_url_path }}/img/logo.png" title="OnionShare">
    <h1>{% if title %}{{ title }}{% else %}OnionShare{% endif %}</h1>
  </header>

  <div class="info-wrapper">
    <div class="info">
      <p><img class="logo" src="{{ static_url_path }}/img/logo_large.png" title="OnionShare"></p>
      <p class="info-header">Thank you for using OnionShare</p>
      <p class="info-description">You may now close this window.</p>
    </div>
  </div>
</body>

</html>