<!DOCTYPE html>
<html>

<head>
  <title>{% if title %}{{ title }}{% else %}OnionShare Chat{% endif %}</title>
  <link href="{{ static_url_path }}/img/favicon.ico" rel="icon" type="image/x-icon">
  <link rel="stylesheet" rel="subresource" type="text/css" href="{{ static_url_path }}/css/style.css" media="all">
</head>

<body>

  <header class="clearfix">
    <img class="logo" src="{{ static_url_path }}/img/logo.png" title="OnionShare">
    <h1>{% if title %}{{ title }}{% else %}OnionShare Chat{% endif %}</h1>
  </header>
  <noscript>
    <p>
      Chat <b>requires JavaScript</b>, so you must set your Tor Browser security
      level to <b>Safer</b> or <b>Standard</b> to join.
    </p>
  </noscript>

  <div class="chat-container no-js">
    <div class="chat-users">
      <div class="editable-username">
        <label for="username">Your username:</label>
        <input id="username" value="{{ username }}" />
        <p id="username-error"></p>
      </div>
      <ul id="user-list">
      </ul>
    </div>
    <div class="chat-wrapper">
      <div id="chat"></div>

      <div class="chat-form">
        <input type="text" id="new-message" name="new-message" placeholder="Type your message" />
      </div>

    </div>
  </div>
  <script src="{{ static_url_path }}/js/jquery-3.5.1.min.js"></script>
  <script src="{{ static_url_path }}/js/socket.io.min.js"></script>
  <script async src="{{ static_url_path }}/js/chat.js"></script>
</body>

</html>
